# Copyright 2009-2010, BlueDynamics Alliance - http://bluedynamics.com
import unittest
import zope.component
import zope.annotation
import zopyx.txng3.core
import cornerstone.soup.tests
from pprint import pprint
from interlude import interact
from zope.testing import doctest
from zope.app.testing.placelesssetup import setUp, tearDown
from zope.configuration.xmlconfig import XMLConfig

optionflags = doctest.NORMALIZE_WHITESPACE | \
              doctest.ELLIPSIS | \
              doctest.REPORT_ONLY_FIRST_FAILURE

TESTFILES = [
    '../soup.txt',
    '../ting.txt',
]

def test_suite():
    setUp()
    XMLConfig('meta.zcml', zope.component)()
    XMLConfig('configure.zcml', zope.annotation)()
    XMLConfig('configure.zcml', zopyx.txng3.core)()
    XMLConfig('configure.zcml', cornerstone.soup.tests)()
    return unittest.TestSuite([
        doctest.DocFileSuite(
            file, 
            optionflags=optionflags,
            globs={'interact': interact,
                   'pprint': pprint},
        ) for file in TESTFILES
    ])

if __name__ == '__main__':                                   #pragma NO COVERAGE
    unittest.main(defaultTest='test_suite')                  #pragma NO COVERAGE


# -*- coding: iso-8859-15 -*-
from zope.schema import Bool, BytesLine, Int, Choice
from zope.i18nmessageid import ZopeMessageFactory as _
from zopyx.txng3.core import config
import zope.catalog.interfaces
import re

class ITingIndexSchema(zope.catalog.interfaces.IAttributeIndex):
    default_fields=BytesLine(
        title=_(u"Default fields"),
        description=_("Look in these fields by default (consider dedicated_storage=True)"),
        required=True,
        default=''
    )
    storage=Choice(
        title=_(u"Storage"),
        description=_(u"Component for storing wordID-to-documentID-mappings"),
        required=True,
        default=config.DEFAULT_STORAGE,
        vocabulary="TextIndexNG3 Storages"
    )
    dedicated_storage=Bool(
        title=_(u"Dedicated storage"),
        description=_(u"Use seperate index for each field (allows per field search)"),
        required=True,
        default=config.defaults['dedicated_storage']
    )
    languages=BytesLine(
        title=_(u"Languages"),
        description=_(u"Languages supported by this index (space seperated list)"),
        default=config.DEFAULT_LANGUAGE,
        constraint=re.compile('[a-z]+(\s+[a-z]+)*').match
    )
    use_stemmer=Bool(
        title=_(u"Stemming"),
        description=_(u"Compare words according to their word stems (a kind of similarity search)"),
        default=False,
        required=config.defaults['use_stemmer']
    )
    use_stopwords=Bool(
        title=_(u"Stopwords"),
        description=_(u"Enable to prvent looking for words like 'and' or 'a' which are unlike to be useful in search queries"),
        required=True,
        default=config.defaults['use_stopwords']
    )
    use_normalizer=Bool(
        title=_(u"Normalize"),
        description=_(u"Enable to normalize words language specific (e.g. Ã€ -> ae , Ãš -> e)"),
        required=True,
        default=config.defaults['use_normalizer']
    )
    ranking=Bool(
        title=_(u"Ranking"),
        description=_(u"Enable ranking according to word frequency of documents (selects different storage)"),
        required=True,
        default=config.defaults['ranking']
    )
    autoexpand_limit=Int(
        title=_(u"Autoexpand limit"),
        description=_(u"Lower limit for automatic right-*-wildcard-search"),
        required=True,
        default=config.defaults['autoexpand_limit']
    )
    splitter=Choice(
        title=_(u"Splitter"),
        description=_(u"Splitter to be used to turn text into words"),
        required=True,
        default=config.DEFAULT_SPLITTER,
        vocabulary="TextIndexNG3 Splitters"
    )
    lexicon=Choice(
        title=_(u"Lexicon"),
        description=_(u"Component to be used for storing word-to-id-mappings"),
        required=True,
        default=config.DEFAULT_LEXICON,
        vocabulary="TextIndexNG3 Lexicons"
    )
    index_unknown_languages=Bool(
        title=_(u"Index unknown languages"),
        description=_(u"Assigns unknown languages the first language of the languages selected for this index"),
        required=True,
        default=config.defaults['index_unknown_languages'],
    )
    splitter_additional_chars=BytesLine(
        title=_(u"Non-seperators"),
        description=_(u"Characters that should *not* be threaded as separators"),
        required=True,
        default=config.defaults['splitter_additional_chars']
    )
    splitter_casefolding=Bool(
        title=_(u"Case-insensitive"),
        description=_(u"Make this index case insensitive"),
        required=True,
        default=config.defaults['splitter_casefolding']
    )
    query_parser=Choice(
        title=_(u"Query Parser"),
        description=_(u"Parser to be used for this index"),
        required=True,
        default=config.DEFAULT_PARSER,
        vocabulary="TextIndexNG3 Query Parsers" 
    )

class ITingIndex(ITingIndexSchema, zope.catalog.interfaces.ICatalogIndex):
    pass


###############################################################################
# * This code has been contribued by Ivo van der Wijk
# * This code was improved by Frank Burkhardt
# * More improvements and tests by Simon Pamies
# * Stripped down dependencies to zope.catalog and txng core stuff by 
#   Robert Niederreiter
###############################################################################

import persistent
import BTrees

import zope.interface
from zope.component import createObject
from zope.index import interfaces
import zope.catalog.attribute
import zope.index.interfaces

from zopyx.txng3.core import config
from zopyx.txng3.core.index import Index
from zopyx.txng3.core.interfaces import IStorageWithTermFrequency

from cornerstone.soup.interfaces.ting import ITingIndex

defaults = {
    'use_stemmer': config.defaults['use_stemmer'],
    'dedicated_storage': config.defaults['dedicated_storage'],
    'ranking': config.defaults['ranking'],
    'use_normalizer': config.defaults['use_normalizer'],
    'languages': config.DEFAULT_LANGUAGE,
    'use_stopwords': config.defaults['use_stopwords'],
    'autoexpand_limit': config.defaults['autoexpand_limit'],
    'splitter': config.DEFAULT_SPLITTER,
    'index_unknown_languages': config.defaults['index_unknown_languages'],
    'query_parser': config.DEFAULT_PARSER,
    'lexicon': config.DEFAULT_LEXICON,
    'splitter_add_chars': config.defaults['splitter_additional_chars'],
    'storage': config.DEFAULT_STORAGE,
    'splitter_casefolding': config.defaults['splitter_casefolding']
}

class TingIndex(zope.catalog.text.TextIndex, 
                persistent.Persistent):

    zope.interface.implements(
        zope.index.interfaces.IInjection,
        zope.index.interfaces.IStatistics,
        zope.index.interfaces.IIndexSearch,
        ITingIndex)
    
    def __init__(self,
                 field_name=None,
                 interface=None,
                 field_callable=False,
                 use_stemmer=defaults['use_stemmer'],
                 dedicated_storage=defaults['dedicated_storage'],
                 ranking=defaults['ranking'],
                 use_normalizer=defaults['use_normalizer'],
                 languages=defaults['languages'],
                 use_stopwords=defaults['use_stopwords'],
                 autoexpand_limit=defaults['autoexpand_limit'],
                 splitter=defaults['splitter'],
                 index_unknown_languages=defaults['index_unknown_languages'],
                 query_parser=defaults['query_parser'],
                 lexicon=defaults['lexicon'],
                 splitter_additional_chars=defaults['splitter_add_chars'],
                 storage=defaults['storage'],
                 splitter_casefolding=defaults['splitter_casefolding'],
                 asIFSet=True):
        if ranking:
            util = createObject(storage)
            if not IStorageWithTermFrequency.providedBy(util):
                raise ValueError("This storage cannot be used for ranking")
        if isinstance(field_name, basestring):
            _fields = field_name.split(' ')
        else:
            _fields = field_name
        zope.catalog.attribute.AttributeIndex.__init__(
            self, _fields[0], interface, field_callable)
        if len(_fields) < 2:
            dedicated_storage = False
        _default_fields = [_fields[0]]
        self._index = Index(
            fields=_fields,
            languages=languages.split(' '),
            use_stemmer=use_stemmer,
            dedicated_storage=dedicated_storage,
            ranking=ranking,
            use_normalizer=use_normalizer,
            use_stopwords=use_stopwords,
            storage=storage,
            autoexpand_limit=autoexpand_limit,
            splitter=splitter,
            lexicon=lexicon,
            index_unknown_languages=index_unknown_languages,
            query_parser=query_parser,
            splitter_additional_chars=splitter_additional_chars,
            splitter_casefolding=splitter_casefolding
        )
        self.languages=languages
        self.use_stemmer=use_stemmer
        self.dedicated_storage=dedicated_storage
        self.ranking=ranking
        self.use_normalizer=use_normalizer
        self.use_stopwords=use_stopwords
        self.interface = interface
        self.storage=storage
        self.autoexpand_limit=autoexpand_limit
        self.default_fields=_default_fields
        self._fields=_fields
        self.splitter=splitter
        self.lexicon=lexicon
        self.index_unknown_languages=index_unknown_languages
        self.query_parser=query_parser
        self.splitter_additional_chars=splitter_additional_chars
        self.splitter_casefolding=splitter_casefolding
        self._asIFSet = asIFSet

    def clear(self):
        self._index.clear()

    def documentCount(self):
        """See interface IStatistics
        """
        return len(self._index.getStorage(self.default_fields[0]))

    def wordCount(self):
        """See interface IStatistics
        """
        return len(self._index.getLexicon())

    def index_doc(self, docid, value):
        """See interface IInjection
        """
        if value is not None:
            self._index.index_object(value, docid)

    def unindex_doc(self, docid):
        """See interface IInjection
        """
        self._index.unindex_object(docid)

    def apply(self, query):
        kw = dict()
        if isinstance(query, dict):
            kw.update(query)
            query = kw['query']
            del kw['query']
        res = self._index.search(query, **kw).getDocids()
        if self._asIFSet:
            return BTrees.IFBTree.IFSet(res)
        return res


# Copyright 2009-2010, BlueDynamics Alliance - http://bluedynamics.com
import uuid
import random
import tempfile
from copy import deepcopy 
from zope.interface import implements
from zope.component import getUtility
from zope.annotation import IAnnotations
from zope.index.interfaces import IIndexSort
from zope.catalog.field import FieldIndex
from Acquisition import (
    aq_inner,
    aq_parent,
    aq_base,
)
from BTrees.IOBTree import IOBTree
from BTrees.OOBTree import OOBTree
from BTrees.Length import Length
from OFS.SimpleItem import SimpleItem
from cornerstone.soup.interfaces import (
    ISoupAnnotatable,
    ISoup,
    IRecord,
    ICatalogFactory
)

CACHE_PREFIX = 'soup_storage_%s'

class NoLongerSupported(Exception): pass

class StorageLocator(object):
    
    def __init__(self, context):
        self.context = context
    
    @property
    def root(self):
        obj = aq_inner(self.context)
        while True:
            if ISoupAnnotatable.providedBy(obj):
                return obj
            obj = aq_parent(obj)
            if not obj:
                raise AttributeError(u"Invalid soup context.")
    
    def storage(self, id):
        return self.locate(id)
    
    def path(self, id):
        annotations = IAnnotations(self.root)
        if isinstance(annotations[id], basestring):
            return annotations[id]
        return '/'
    
    def traverse(self, path):
        obj = self.root
        path = path.split('/')
        for name in path:
            try:
                obj = obj[name]
            except AttributeError, e:
                msg = u'Object at %s does not exist.' % '/'.join(path)
                raise ValueError(msg)
        return obj
    
    def annotated(self, obj, id):
        annotations = IAnnotations(obj)
        if not id in annotations:
            annotations[id] = SoupData()
        return annotations[id]
    
    def locate(self, id):
        entity = self.annotated(self.root, id)
        if isinstance(entity, basestring):
            entity = self.annotated(self.traverse(entity), id)
        if isinstance(entity, SoupData):
            return entity
        msg = u'Conflicting annotation with name %s. No SoupData contained'
        msg = msg % id
        raise ValueError(msg)
            
    def _invalidate_cache(self, id):
        key = CACHE_PREFIX % id
        if self.context.REQUEST.get(key):
            del self.context.REQUEST[key]
        
    def move(self, sid, target_path):        
        target_context = self.traverse(target_path)
        target_annotations = IAnnotations(target_context)        
        if sid in target_annotations:
            raise KeyError, 'Annotation-Key %s already used at %s' % \
                            (sid, target_path)
        root_annotations = IAnnotations(self.root)
        source_data = self.storage(sid).data
        root_ann_obj = root_annotations[sid]
        if not isinstance(root_ann_obj, SoupData):            
            source_annotations = IAnnotations(self.traverse(root_ann_obj))
            del source_annotations[sid]
        root_annotations[sid] = target_path
        self._invalidate_cache(sid)
        target_soup = getSoup(self.context, sid)
        # access soup.storage once if empty soup is copied. annotation is
        # created on first storage access !!!
        target_soup.storage
        for key in source_data:
            target_soup.add(deepcopy(source_data[key]))            
    
    def relocate(self, sid, target_path):
        target_context = self.traverse(target_path)
        target_annotations = IAnnotations(target_context)        
        if sid not in target_annotations:
            raise KeyError, 'Annotation-Key %s must exist at %s' % \
                            (sid, target_path)
        root_annotations = IAnnotations(self.root)
        root_annotations[sid] = target_path
        self._invalidate_cache(sid)
        target_soup = getSoup(self.context, sid)        
        
class SoupData(SimpleItem):
    
    def __init__(self):
        self.data = IOBTree()
        self.catalog = None
        self._length = Length()
        
    @property
    def length(self):
        if not hasattr(self, '_length'):
            # BBB
            self._length = Length()
        return self._length
        
    def __len__(self):
        return self.length()

def getSoup(context, id):
    return Soup(context, id)

#class Soup(object): XXX TODO: for version 3.0 make Soup simple object derived
class Soup(SimpleItem):
    implements(ISoup)
    
    def __init__(self, context, id):
        self.context = context
        self.id = id
    
    @property
    def _cachekey(self):
        return CACHE_PREFIX % self.id
    
    @property
    def storage(self):
        if not hasattr(self, 'context'):
            error = u'Using ISoup as utility is no longer supported. ' + \
                    u'Register your Soup as adapter instead and change the ' + \
                    u'soup lookups in your code.'
            raise NoLongerSupported(error)
        request = self.context.REQUEST
        cachekey = self._cachekey
        if request.get(cachekey):
            return request[cachekey]
        storage = StorageLocator(self.context).storage(self.id)
        request[cachekey] = storage
        return storage
    
    @property
    def data(self):
        return self.storage.data
    
    @property
    def catalog(self):
        storage = self.storage
        if not storage.catalog:
            storage.catalog = getUtility(ICatalogFactory, name=self.id)()
        return storage.catalog

    def add(self, record):
        record.intid = self._generateid()
        self.data[record.intid] = record
        self.storage.length.change(1)
        record = self.data[record.intid] # ?
        self.catalog.index_doc(record.intid, record)
        # XXX: notify subscribers here if not done by OFS, check this.
        return record.intid
    
    def _query(self, **kw):
        sort_index = kw.pop('_sort_index', None)
        limit = kw.pop('_limit', None)
        reverse = kw.pop('_reverse', False)
        catalog = self.catalog
        querykw = {}
        for key in kw:
            # XXX: remove special case in later version
            if isinstance(catalog[key], FieldIndex) \
              and not isinstance(kw[key], tuple) \
              and not isinstance(kw[key], list):
                querykw[key] = (kw[key], kw[key])
            else:
                querykw[key] = kw[key]
        results = catalog.apply(querykw)
        if results is not None:
            if sort_index is not None:
                index = catalog[sort_index]
                if not IIndexSort.providedBy(index):
                    msg = 'Index %s does not support sorting.' % sort_index
                    raise ValueError(msg)
                results = list(index.sort(results,
                                          limit=limit,
                                          reverse=reverse))
        return results
    
    def query(self, **kw):
        ids = self._query(**kw)
        for id in ids:
            yield self.data[id]
    
    def lazy(self, **kw):
        ids = self._query(**kw)
        for id in ids:
            yield LazyRecord(id, self)
    
    def clear(self):
        self.storage.__init__()
        request = self.context.REQUEST
        cachekey = self._cachekey
        del request[cachekey]
        self.rebuild()
    
    def rebuild(self):
        self.storage.catalog = getUtility(ICatalogFactory, name=self.id)()
        self.reindex()
    
    def reindex(self, records=None):
        if records is None:
            records = self.data.values()
        for record in records:
            self.catalog.index_doc(record.intid, record)

    def __delitem__(self, record):
        try:
            del self.data[record.intid]
            self.storage.length.change(-1)
        except Exception, e:
            raise e
        self.catalog.unindex_doc(record.intid)
    
    _v_nextid = None
    _randrange = random.randrange
    
    def _generateid(self):
        # Stolen from zope.app.intid.
        while True:
            if self._v_nextid is None:
                self._v_nextid = self._randrange(0, 2**31)
            uid = self._v_nextid
            self._v_nextid += 1
            if uid not in self.data:
                return uid
            self._v_nextid = None

class LazyRecord(object):
    
    def __init__(self, intid, soup):
        self.intid = intid
        self.soup = soup
    
    def __call__(self):
        return self.soup.data[self.intid]

EMPTY_MARKER = object()

class Record(SimpleItem):
    implements(IRecord)
    
    def __init__(self, **kw):
        self.id = uuid.uuid4().hex
        self.intid = None
        self.data = OOBTree()
        for key in kw.keys():
            self.data[key] = kw[key]
        self._p_changed = True
    
    def __getattribute__(self, name):
        try:
            attr = SimpleItem.__getattribute__(self, 'data').get(name,
                                                                 EMPTY_MARKER)
            if attr is not EMPTY_MARKER:
                return attr
        except AttributeError, e:
            pass
        return SimpleItem.__getattribute__(self, name)
        
    def __deepcopy__(self, memo):
        rec = self.__class__()
        rec.id = self.id
        rec.intid = self.intid
        for key in self.data:
            rec.data[key] = deepcopy(self.data[key])
        return rec             


# Copyright 2010, BlueDynamics Alliance - http://bluedynamics.com
import transaction
from Acquisition import aq_inner
from zope.component import (
    getUtility,
    getUtilitiesFor,
)
from zope.component.interfaces import ComponentLookupError
from cornerstone.soup.interfaces import (
    ISoup,
    ICatalogFactory,
)
from cornerstone.soup import (
    getSoup,
    StorageLocator,
)
from Products.Five import BrowserView

class SoupAdmin(BrowserView):
    
    @property
    def info(self):
        return self.request.form.get('info', '')
        
    @property
    def existing_soups(self):
        ret = list()
        for name, soup in getUtilitiesFor(ICatalogFactory):
            ret.append(name)
        return ret
    
    def count(self, soup):
        soup = getSoup(aq_inner(self.context), soup)
        return len(soup.storage)
    
    def redirect_base(self, msg):
        url = self.context.absolute_url()
        url += '/soup-controlpanel?info=%s' % msg
        self.request.RESPONSE.redirect(url)
    
    def storage_path(self, id):        
        try:
            path = StorageLocator(self.context).path(id)
        except KeyError, e:
            path = '/ (not initialized yet)'
        else:
            if not path.startswith('/'):
                path = '/' + path
        return path 
    
    def storage_migrated(self, id):
        try:
            soup = getUtility(ISoup, name=id)
            soup.context = self.context
            if 'data' in soup.__dict__:
                return False
        except ComponentLookupError, e:
            pass
        return True
    
    def local_component_removed(self, id):
        try:
            soup = getUtility(ISoup, name=id)
            return False
        except ComponentLookupError, e:
            pass
        return True
    
    def reindex_soup(self):
        id = self.request.form.get('id')
        if not id:
            return self.redirect_base('No id')
        soup = getSoup(self.context, id)
        soup.reindex()
        msg = '%s reindexed.' % id
        return self.redirect_base(msg)
    
    def rebuild_soup(self):
        id = self.request.form.get('id')
        if not id:
            return self.redirect_base(msg)
        soup = getSoup(self.context, id)
        soup.rebuild()
        msg = '%s rebuilt.' % id
        return self.redirect_base(msg)
    
    def migrate_storage(self):
        id = self.request.form.get('id')
        if not id:
            return self.redirect_base(u'No id given')
        soup = getUtility(ISoup, name=id)
        soup.context = self.context
        if 'data' in soup.__dict__:
            soup.storage.data = soup.__dict__['data']
            del soup.__dict__['data']
            if '_catalog' in soup.__dict__:
                del soup.__dict__['_catalog']
            soup._p_changed = True
            soup.rebuild()
            msg = '%s migrated.' % id
            return self.redirect_base(msg)
        msg = '%s already up to date.' % id
        return self.redirect_base(msg)
    
    def remove_local_component(self):
        id = self.request.form.get('id')
        if not id:
            return self.redirect_base(u'No id given')
        soup = getUtility(ISoup, name=id)
        root = StorageLocator(self.context).root
        root.getSiteManager().unregisterUtility(soup, ISoup, id)
        del soup
        transaction.commit()
        msg = '%s local component removed.' % id
        return self.redirect_base(msg)
    
    def move_storage(self):
        id = self.request.form.get('id')
        if not id:
            return self.redirect_base(u'No id given')
        path = self.request.form.get('path')
        if not path:
            return self.redirect_base(u'No path given')
        locator = StorageLocator(self.context)
        try:
            target = locator.traverse(path)
        except ValueError:
            return self.redirect_base(u'Desired path not exists')
        method = self.request.form.get('moveormount')
        if method == 'move':
            locator.move(id, path)
            transaction.commit()
            return self.redirect_base(u'Moved storage of %s to %s' % (id, path))
        elif method == "mount":
            locator.relocate(id, path)
            transaction.commit()
            return self.redirect_base(u'Mounted storage %s to %s' % (id, path))
        else:
            return self.redirect_base(u'Invalid action (move or mount only)')
    
    def rebuild_length(self):
        id = self.request.form.get('id')
        if not id:
            return self.redirect_base(msg)
        soup = getSoup(self.context, id)
        newlen = len(soup.storage.data)
        soup.storage.length.set(newlen)
        transaction.commit()
        return self.redirect_base(u'Length of storage %s is %s' % (id, newlen))        


# Copyright 2009-2010, BlueDynamics Alliance - http://bluedynamics.com
from zope.interface import implements
from zope.catalog.catalog import Catalog
from zope.catalog.field import FieldIndex
from zope.catalog.text import TextIndex
from zope.catalog.keyword import KeywordIndex
from cornerstone.soup.interfaces import ICatalogFactory
from cornerstone.soup.ting import TingIndex

class MyCatalogFactory(object):
    """ICatalogFactory implementation used for testing.
    """
    implements(ICatalogFactory)
    
    def create_catalog(self):
        catalog = Catalog()
        catalog[u'user'] = FieldIndex(field_name='user', field_callable=False)
        catalog[u'text'] = TextIndex(field_name='text', field_callable=False)
        catalog[u'keywords'] = KeywordIndex(field_name='keywords',
                                            field_callable=False)
        return catalog

    def __call__(self):
        self.catalog = self.create_catalog()
        return self.catalog

class TingCatalogFactory(object):
    """ICatalogFactory implementation for testing textindex NG integration to
    soup.
    """
    implements(ICatalogFactory)
    
    def create_catalog(self):
        catalog = Catalog()
        catalog[u'ting'] = TingIndex(field_name=('foo', 'bar', 'baz'),
                                     field_callable=False)
        return catalog
    
    def __call__(self):
        self.catalog = self.create_catalog()
        return self.catalog

class SortCatalogFactory(object):
    """ICatalogFactory implementation for testing sorting in soup.
    """
    implements(ICatalogFactory)
    
    def __call__(self):
        catalog = Catalog()
        catalog[u'date'] = FieldIndex(field_name='date',
                                      field_callable=False)
        return catalog